package checkers.javari;

import javax.lang.model.element.*;
import com.sun.source.tree.*;

import checkers.source.*;
import checkers.basetype.*;
import checkers.javari.quals.Assignable;
import checkers.types.*;
import checkers.types.AnnotatedTypeMirror.*;
import checkers.util.TreeUtils;


/**
 * A type-checking visitor for the Javari mutability annotations
 * ({@code @ReadOnly}, {@code @Mutable} and {@code @Assignable}) that
 * extends BaseTypeVisitor.
 *
 * @see BaseTypeVisitor
 */
public class JavariVisitor extends BaseTypeVisitor<Void, Void> {

    final private AnnotationMirror READONLY, MUTABLE, POLYREAD, QREADONLY;

    /**
     * Creates a new visitor for type-checking the Javari mutability
     * annotations.
     *
     * @param checker the {@link JavariChecker} to use
     * @param root the root of the input program's AST to check
     */
    public JavariVisitor(JavariChecker checker, CompilationUnitTree root) {
        super(checker, root);
        READONLY = checker.READONLY;
        MUTABLE = checker.MUTABLE;
        POLYREAD = checker.POLYREAD;
        QREADONLY = checker.QREADONLY;
    }

    /**
     * Ensures the class type is not {@code @PolyRead} outside a
     * {@code @PolyRead} context.
     */
    @Override
    public Void visitClass(ClassTree node, Void p) {
        if (atypeFactory.fromClass(node).hasAnnotation(POLYREAD)
            && !atypeFactory.getSelfType(node).hasAnnotation(POLYREAD))
            checker.report(Result.failure("polyread.type"), node);

        return super.visitClass(node, p);
    }

   /**
    * Checks whether the variable represented by the given type and
    * tree can be assigned, causing a checker error otherwise.
    */
    @Override
    protected void checkAssignability(AnnotatedTypeMirror varType, Tree varTree) {
        if (!(varTree instanceof ExpressionTree)) return;
        Element varElt = varType.getElement();
        if (varElt != null && varElt.getAnnotation(Assignable.class) != null)
            return;

        boolean variableLocalField =
            TreeUtils.isSelfAccess((ExpressionTree) varTree)
            && varElt != null
            && varElt.getKind().isField();
        boolean inConstructor =  TreeUtils.isConstructor(visitorState.getMethodTree());
        AnnotatedDeclaredType mReceiver = atypeFactory.getSelfType(varTree);

        if (variableLocalField && !inConstructor && !mReceiver.hasAnnotation(MUTABLE))
            checker.report(Result.failure("ro.field"), varTree);

        if (varTree.getKind() == Tree.Kind.MEMBER_SELECT) {
            AnnotatedTypeMirror receiver = atypeFactory.getReceiver((ExpressionTree)varTree);
            if (receiver != null && !receiver.hasAnnotation(MUTABLE))
                checker.report(Result.failure("ro.field"), varTree);
        }
    }


    /**
     * Tests whether the tree expressed by the passed type tree
     * contains a qualified primitive type on its qualified type, and
     * if so emits an error.
     *
     * @param tree  the AST type supplied by the user
     */
    @Override
    public void validateTypeOf(Tree tree) {
        Tree tTree;
        if (tree.getKind() == Tree.Kind.VARIABLE) {
            tTree = ((VariableTree)tree).getType();
            switch (tTree.getKind()) {
            case PRIMITIVE_TYPE: {
                AnnotatedTypeMirror vType = atypeFactory.getAnnotatedType(tree);
                if (vType.hasAnnotation(QREADONLY)
                    || vType.hasAnnotation(READONLY)
                    || vType.hasAnnotation(POLYREAD))
                    checker.report(Result.failure("primitive.ro"), tree);
                break;
            }
            case PARAMETERIZED_TYPE: {
                ParameterizedTypeTree pt = (ParameterizedTypeTree)tTree;
                for (Tree t : pt.getTypeArguments()) validateTypeOf(t);
                break;
            }
            }
        } else
            tTree = tree;

        switch (tTree.getKind()) {
        case ANNOTATED_TYPE: {
            AnnotatedTypeMirror vType = atypeFactory.getAnnotatedType(tree);
            Tree nextTree = ((AnnotatedTypeTree)tTree).getUnderlyingType();
            if (nextTree.getKind() == Tree.Kind.PRIMITIVE_TYPE) {
                if (vType.hasAnnotation(QREADONLY)
                    || vType.hasAnnotation(READONLY)
                    || vType.hasAnnotation(POLYREAD)) {
                    checker.report(Result.failure("primitive.ro"), tTree);
                }
            } else {
                validateTypeOf(nextTree);
            }
            break;
        }
        case ARRAY_TYPE: {
            validateTypeOf(((ArrayTypeTree)tTree).getType());
            break;
        }
        }
        super.validateTypeOf(tree);
    }


    /**
     * Emits a warning if a cast is made increasing mutability.
     */
    @Override
    public Void visitTypeCast(TypeCastTree node, Void p) {
        AnnotatedTypeMirror exprType = atypeFactory.getAnnotatedType(node.getExpression()),
            castType = atypeFactory.getAnnotatedType(node);

        if (!checker.isSubtype(exprType, castType)) {
            checker.report(Result.warning("mutable.cast"), node.getType());
        }

        return super.visitTypeCast(node, p);
    }
}
